from ..utils import assert_type
from . import wire_pb2
from enum import Enum
from six import string_types


class StatusCode(Enum):
    UNKNOWN = wire_pb2.StatusCode.Value('UNKNOWN')
    OK = wire_pb2.StatusCode.Value('OK')
    CANCELLED = wire_pb2.StatusCode.Value('CANCELLED')
    INVALID_ARGUMENT = wire_pb2.StatusCode.Value('INVALID_ARGUMENT')
    DEADLINE_EXCEEDED = wire_pb2.StatusCode.Value('DEADLINE_EXCEEDED')
    NOT_FOUND = wire_pb2.StatusCode.Value('NOT_FOUND')
    ALREADY_EXISTS = wire_pb2.StatusCode.Value('ALREADY_EXISTS')
    PERMISSION_DENIED = wire_pb2.StatusCode.Value('PERMISSION_DENIED')
    UNAUTHENTICATED = wire_pb2.StatusCode.Value('UNAUTHENTICATED')
    FAILED_PRECONDITION = wire_pb2.StatusCode.Value('FAILED_PRECONDITION')
    OUT_OF_RANGE = wire_pb2.StatusCode.Value('OUT_OF_RANGE')
    UNIMPLEMENTED = wire_pb2.StatusCode.Value('UNIMPLEMENTED')
    INTERNAL_ERROR = wire_pb2.StatusCode.Value('INTERNAL_ERROR')


class Status(object):
    def __init__(self, code=StatusCode.UNKNOWN, why=""):
        self._code = None
        self._why = None
        self.code = code
        self.why = why

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __str__(self):
        pretty = "{{ code={} why='{}' }}".format(self.code, self.why or "")
        return pretty

    @property
    def code(self):
        return self._code

    @code.setter
    def code(self, code):
        assert_type(code, StatusCode, "code")
        self._code = code

    def ok(self):
        return self._code == StatusCode.OK

    @property
    def why(self):
        return self._why

    @why.setter
    def why(self, reason):
        assert_type(reason, string_types, "why")
        self._why = reason



from datetime import datetime
from google.protobuf import json_format
from six import integer_types, string_types, binary_type

from .utils import now, assert_type, new_uuid
from .subscription import Subscription
from .wire.status import Status
from .wire.content_type import ContentType
from .tracing.propagation import TextFormatPropagator


class Message(object):
    def __init__(self, content=None, reply_to=None, content_type=None):
        """ Creates a new message.
        Args:
            content (str or object): sets the message body with the
            given content. If an object is provided, it will be packed
            using the given content_type.

            content_type (ContentType): describes how the content will
            be serialized.

            reply_to (str or Subscription): indicates where the reply to
            this message should be sent to.
        """
        self._topic = None
        self._body = ''
        self._reply_to = None
        self._subscription_id = None
        self._correlation_id = None
        self._content_type = None
        self._created_at = now()
        self._metadata = {}
        self._timeout = None
        self._status = None

        if reply_to is not None:
            self.reply_to = reply_to

        if content_type is not None:
            self.content_type = content_type

        if content is not None:
            if isinstance(content, binary_type):
                self.body = content
            else:
                self.pack(content)

    def __str__(self):
        """ Converts a message to a verbose string of its properties """
        created_at = datetime.fromtimestamp(self.created_at)
        pretty = "{\n"
        pretty += "  topic = '{}'\n".format(self.topic or "")
        pretty += "  created_at = {}\n".format(created_at)
        pretty += "  correlation_id = {}\n".format(self.correlation_id)
        pretty += "  reply_to = '{}'\n".format(self.reply_to or "")
        pretty += "  subscription_id = '{}'\n".format(self.subscription_id
                                                      or "")
        pretty += "  timeout = {}\n".format(self.timeout)
        pretty += "  status = {}\n".format(self.status)
        pretty += "  metadata = {}\n".format(self.metadata)
        pretty += "  content_type = {}\n".format(self.content_type)
        pretty += "  body[{}] = {} \n".format(len(self.body), repr(self.body))
        pretty += "}"
        return pretty

    def short_string(self):
        """ Converts a message to a simplified string of its properties, empty
         fields are not printed """
        created_at = datetime.fromtimestamp(self.created_at)
        pretty = "{"
        pretty += "topic='{}'".format(self.topic)
        pretty += " created_at={}".format(created_at)
        if self.has_correlation_id():
            pretty += " correlation_id={}".format(self.correlation_id)
        if self.has_reply_to():
            pretty += " reply_to='{}'".format(self.reply_to)
        if self.has_subscription_id():
            pretty += " subscription_id='{}'".format(self.subscription_id)
        if self.has_timeout():
            pretty += " timeout={}".format(self.timeout)
        if self.has_status():
            pretty += " status={}".format(self.status)
        if self.has_metadata():
            pretty += " metadata={}".format(self.metadata)
        if self.has_content_type():
            pretty += " content_type={}".format(self.content_type)
        pretty += " body[{}]={}".format(len(self.body), repr(self.body))
        pretty += "}"
        return pretty

    def __eq__(self, other):
        """ Returns True if the messages are equal, False otherwise """
        return self.__dict__ == other.__dict__

    def create_reply(self):
        reply = Message()
        if self.has_reply_to():
            reply.topic = self.reply_to
        if self.has_correlation_id():
            reply.correlation_id = self.correlation_id
        if self.has_content_type():
            reply.content_type = self.content_type
        return reply

    # topic

    @property
    def topic(self):
        """ str: Topic where the message was published or is going
         to be published """
        return self._topic

    @topic.setter
    def topic(self, topic):
        assert_type(topic, string_types, "topic")
        self._topic = topic

    def has_topic(self):
        """ Returns: True if the property topic of the message is set,
         False otherwise """
        return bool(self._topic)

    # reply_to

    @property
    def reply_to(self):
        """ str: Topic where the reply to this message should be published.
         When setting this property an object of type Subscription can be
         passed to automatically set this value. The correlation_id field
         is automatically set if empty. """
        return self._reply_to

    @reply_to.setter
    def reply_to(self, value):
        assert_type(value, list(string_types) + [Subscription], "reply_to")

        if self.correlation_id is None:
            self.correlation_id = new_uuid()

        if isinstance(value, Subscription):
            self._reply_to = value.name
            self.subscription_id = value.id

        elif isinstance(value, string_types):
            self._reply_to = value

    def has_reply_to(self):
        """ Returns: True if the property reply_to of the message is set,
         False otherwise """
        return bool(self._reply_to)

    # subscription_id

    @property
    def subscription_id(self):
        """ str: ID of the subscription that this message belongs to """
        return self._subscription_id

    @subscription_id.setter
    def subscription_id(self, value):
        assert_type(value, string_types, "subscription_id")
        self._subscription_id = value

    def has_subscription_id(self):
        """ Returns: True if the property subscription_id of the message is set,
         False otherwise """
        return bool(self._subscription_id)

    # correlation_id

    @property
    def correlation_id(self):
        """ int: Unique ID used to correlate reply/response messages """
        return self._correlation_id

    @correlation_id.setter
    def correlation_id(self, value):
        assert_type(value, integer_types, "correlation_id")
        self._correlation_id = value

    def has_correlation_id(self):
        """ Returns: True if the property correlation_id of the message is set,
         False otherwise """
        return self._correlation_id is not None

    # body

    @property
    def body(self):
        """ bytes: Raw content of the message """
        return self._body

    @body.setter
    def body(self, value):
        assert_type(value, binary_type, "body")
        self._body = value

    def has_body(self):
        """ Returns: True if the property body of the message is set,
         False otherwise """
        return bool(self._body)

    # content_type

    @property
    def content_type(self):
        """ ContentType: Indicates how the content/body of the message
        was serialized """
        return self._content_type

    @content_type.setter
    def content_type(self, value):
        assert_type(value, ContentType, "content_type")
        self._content_type = value

    def has_content_type(self):
        """ Returns: True if the property content_type of the message is set,
         False otherwise """
        return self._content_type is not None

    # created_at

    @property
    def created_at(self):
        """ float: Seconds since the epoch indicating when the message
        was created"""
        return self._created_at

    @created_at.setter
    def created_at(self, timestamp):
        # assert_type(timestamp, (), "created_at")
        self._created_at = timestamp

    def has_created_at(self):
        """ Returns: True if the property created_at of the message is set,
         False otherwise """
        return self._created_at is not None

    # metadata

    @property
    def metadata(self):
        """ dict: Key-value pairs which any can model any type of extra
        information about the message """
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        assert_type(value, dict, "metadata")
        self._metadata = value

    def has_metadata(self):
        """ Returns: True if the property metadata of the message is set,
         False otherwise """
        return len(self._metadata) != 0

    # timeout

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, seconds):
        assert_type(seconds, [float] + list(integer_types), "timeout")
        self._timeout = seconds

    def has_timeout(self):
        """ Returns: True if the property timeout of the message is set,
         False otherwise """
        return self._timeout is not None

    def deadline_exceeded(self):
        if not self.has_timeout():
            return False
        return now() > self.created_at + self.timeout

    # status

    @property
    def status(self):
        """ Status representing the success or not of a RPC """
        return self._status

    @status.setter
    def status(self, value):
        assert_type(value, Status, "status")
        self._status = value

    def has_status(self):
        """ Returns: True if the property status of the message is set,
         False otherwise """
        return self._status is not None

    # tracing

    def extract_tracing(self):
        return TextFormatPropagator.from_carrier(self.metadata)

    def inject_tracing(self, span):
        span_context = TextFormatPropagator.new_span_context(
            trace_id=span.context_tracer.trace_id,
            span_id=span.span_id,
        )
        self.metadata = TextFormatPropagator.to_carrier(
            span_context, self.metadata)

    # pack / unpack

    def pack(self, obj):
        """ Serializes the given object using the specified message
        content_type. If the message has no content_type, the protobuf
        format is used.
        Args:
            obj (object): protobuf object to be serialized.
        """
        if not self.has_content_type():
            self.content_type = ContentType.PROTOBUF

        if self.content_type == ContentType.PROTOBUF:
            # SerializeToString returns py2: str, py3: bytes
            self.body = obj.SerializeToString()

        elif self.content_type == ContentType.JSON:
            # MessageToJson returns py2: str, py3: str
            packed = json_format.MessageToJson(
                obj, indent=0, including_default_value_fields=True)
            if not isinstance(packed, binary_type):
                self.body = packed.encode('latin')
            else:
                self.body = packed
        else:
            raise NotImplementedError(
                "Serialization to '{}' type not implemented".format(
                    self.content_type.name))

    def unpack(self, schema):
        """ Deserializes the content of the message using the given schema.
        If the message has no content_type, the protobuf format is used.
        Args:
            schema (type): type of the protobuf object to be deserialized.
        Returns:
            schema: deserialized instance of the object of type schema.
        """
        obj = schema()
        if not self.has_content_type():
            self.content_type = ContentType.PROTOBUF

        if self.content_type == ContentType.PROTOBUF:
            obj.ParseFromString(self.body)

        elif self.content_type == ContentType.JSON:
            obj = json_format.Parse(self.body, obj)

        else:
            raise NotImplementedError(
                "Deserialization from '{}' type not implemented".format(
                    self.content_type.name))

        return obj



from prometheus_client import start_http_server, Counter
from ..rpc import Interceptor
from ..core import Logger
import time


class MetricsInterceptor(Interceptor):
    def __init__(self):
        self.log = Logger(name='MetricsInterceptor')
        labels = ("service", "status_code")
        self.duration = Counter("rpc_duration_total",
                                "How long requests took in seconds", labels)
        self.count = Counter("rpc_count_total",
                             "How many requests were processed", labels)


    def start_server(self, port=8000):
        start_http_server(port)

    def before_call(self, context):
        self.begin = time.time()

    def after_call(self, context):
        took = time.time() - self.begin
        topic = context.request.topic
        code = context.reply.status.code.name
        self.duration.labels(topic, code).inc(took)
        self.count.labels(topic, code).inc()



from .utils import consumer_id


class Subscription(object):
    def __init__(self, channel, name=None):
        self._id = consumer_id()
        self._name = self._id if name is None else name
        self._topics = set()

        self._channel = channel._channel
        self._exchange = channel._exchange

        def create_queue(channel, exchange, queue, tag, on_message):
            channel.queue_declare(
                queue=queue,
                passive=False,
                durable=False,
                exclusive=False,
                auto_delete=True,
            )

            channel.queue_bind(
                queue=queue,
                exchange=exchange,
                routing_key=queue,
            )

            channel.basic_consume(
                queue=queue,
                callback=on_message,
                consumer_tag=tag,
                no_local=False,
                no_ack=True,
                exclusive=False,
            )

        create_queue(self._channel, self._exchange, self._name, self._id,
                     channel._on_message)

    def subscribe(self, topic):
        """ Subscribes to the given topic.
        Args:
            topic (str): topic to receive messages from.
        """
        self._channel.queue_bind(
            queue=self._name,
            exchange=self._exchange,
            routing_key=topic,
        )
        self._topics.add(topic)

    def unsubscribe(self, topic):
        """ Unsubscribe from the given topic.
        Args:
            topic (str): topic to stop receiving messages from.
        """
        self._channel.queue_unbind(
            queue=self._name,
            exchange=self._exchange,
            routing_key=topic,
        )
        self._topics.remove(topic)

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def topics(self):
        return self._topics



from ..core import Logger, now, StatusCode
from ..rpc import Interceptor


class LogInterceptor(Interceptor):
    def __init__(self):
        self.log = Logger(name='LogInterceptor')

    def before_call(self, context):
        self.begin = now()

    def after_call(self, context):
        took = now() - self.begin
        status = context.reply.status
        if status.ok():
            self.log.info("took={}s, code={}", took, status.code.name)
        elif status.code == StatusCode.INTERNAL_ERROR:
            self.log.error('took={}s status={} request={}', took, status,
                           context.request.short_string())
        else:
            self.log.warn('took={}s status={} request={}', took, status,
                          context.request.short_string())



from opencensus.trace.span_context import SpanContext
from ..utils import assert_type

# Patch opencensus to allow 64 bit trace ids
import opencensus.trace.span_context
import re
opencensus.trace.span_context.TRACE_ID_PATTERN = re.compile('[0-9a-f]{16}?')
opencensus.trace.span_context._INVALID_TRACE_ID = '0' * 16


class TextFormatPropagator(object):
    trace_prefix = 'x-b3'
    trace_id_key = '{}-traceid'.format(trace_prefix)
    span_id_key = '{}-spanid'.format(trace_prefix)
    parent_span_id_key = '{}-parentspanid'.format(trace_prefix)
    sampled_key = '{}-sampled'.format(trace_prefix)
    flags_key = '{}-flags'.format(trace_prefix)

    @classmethod
    def from_carrier(cls, carrier):
        if cls.trace_id_key not in carrier or \
           cls.span_id_key not in carrier:
            return None

        trace_id = carrier[cls.trace_id_key]
        span_id = carrier[cls.span_id_key]

        return SpanContext(
            trace_id=trace_id,
            span_id=span_id,
            from_header=True,
        )

    @classmethod
    def to_carrier(cls, span_context, carrier):
        assert_type(span_context, SpanContext, "span_context")
        carrier[cls.trace_id_key] = span_context.trace_id

        if span_context.span_id is not None:
            carrier[cls.span_id_key] = span_context.span_id

        carrier[cls.sampled_key] = '1'
        carrier[cls.parent_span_id_key] = '0' * 16
        carrier[cls.flags_key] = '0'
        return carrier

    @staticmethod
    def new_span_context(trace_id, span_id):
        return SpanContext(trace_id=trace_id, span_id=span_id)



import os
import uuid
from datetime import datetime


def new_uuid():
    return uuid.uuid4().int >> 64


def consumer_id():
    return '{}/{:X}'.format(os.uname()[1], new_uuid())


def now():
    return (datetime.utcnow() - datetime(1970, 1, 1)).total_seconds()


def assert_type(instance, types, name):
    if isinstance(types, list):
        types = tuple(types)

    if not isinstance(instance, types):
        input_type = type(instance).__name__
        if isinstance(types, tuple):
            types = " or ".join([t.__name__ for t in types])
            error = "Object {} must be of types {}, received type {}".format(
                name, types, input_type)
        else:
            error = "Object {} must be of type {}, received type {}".format(
                name, types.__name__, input_type)
        raise TypeError(error)



from colorlog import ColoredFormatter, StreamHandler, getLogger
from sys import exit
import logging
from .utils import assert_type


class Logger:
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARN = logging.WARN
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    def __init__(self, name, level=logging.DEBUG):
        assert_type(name, str, "name")

        style = "%(log_color)s[%(levelname)-.1s][%(threadName)s]" \
                "[%(asctime)s][%(name)s] %(message)s"

        formatter = ColoredFormatter(
            style,
            datefmt=None,
            reset=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'white',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'white,bg_red',
            },
            secondary_log_colors={},
            style='%')

        self.logger = getLogger(name)
        if len(self.logger.handlers) == 0 and name:
            handler = StreamHandler()
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.set_level(level)

    def set_level(self, level):
        self.logger.setLevel(level)

    def debug(self, formatter, *args):
        self.logger.debug(formatter.format(*args))

    def info(self, formatter, *args):
        self.logger.info(formatter.format(*args))

    def warn(self, formatter, *args):
        self.logger.warning(formatter.format(*args))

    def error(self, formatter, *args):
        self.logger.error(formatter.format(*args))

    def critical(self, formatter, *args):
        self.logger.critical(formatter.format(*args))
        exit(-1)



import amqp
from six.moves import urllib
from .wire.conversion import WireV1


class Channel(object):
    def __init__(self, uri="amqp://guest:guest@localhost:5672"):
        url = urllib.parse.urlparse(uri)

        self.connection = amqp.Connection(
            host="{}:{}".format(url.hostname or "localhost", url.port or 5672),
            userid=url.username or "guest",
            password=url.password or "guest",
            connect_timeout=5.0,
        )
        self.connection.connect()

        self._channel = self.connection.channel()
        self._channel.auto_decode = False

        self._exchange = "is"
        self._channel.exchange_declare(
            exchange=self._exchange,
            type="topic",
            durable=False,
            auto_delete=False,
        )

        self.subscriptions = []
        self.amqp_message = None

    def _on_message(self, message):
        self.amqp_message = message

    def publish(self, message, topic=None):
        """ Publishes a message to the given topic. The topic on the message
        is used when no topic is passed to this function. If no valid topic is
        passed a RuntimeError is raised."""
        if not message.has_topic() and not topic:
            raise RuntimeError("Trying to publish message without topic")

        amqp_message = amqp.Message(
            body=message.body,
            channel=self._channel,
            **WireV1.to_amqp_properties(message))

        self._channel.basic_publish(
            amqp_message,
            exchange=self._exchange,
            routing_key=message.topic if topic is None else topic,
            immediate=False,
            mandatory=False,
        )

    def consume(self, timeout=None):
        """ Blocks waiting for a new message to arrive. If no timeout
        (in seconds) is provided the function blocks forever.
        Args:
            timeout (float): Period in seconds to block waiting for messages.
            Must be a positive number. If 0 is passed the call is non-blocking.
            If no message was received after this period a socket.timeout
            Exception is raised.
        Returns:
            Message: Received message.
        """
        if timeout is not None:
            assert timeout >= 0.0

        self.amqp_message = None
        while self.amqp_message is None:
            self.connection.drain_events(timeout)
        return WireV1.from_amqp_message(self.amqp_message)

    def close(self):
        self.connection.close()



from enum import Enum
from . import wire_pb2
from ..utils import assert_type


class ContentType(Enum):
    PROTOBUF = wire_pb2.ContentType.Value("PROTOBUF")
    JSON = wire_pb2.ContentType.Value("JSON")


def content_type_to_wire(content_type):
    """ Converts an object of type ContentType to the wire string representation.
    Args:
        content_type (ContentType): enum value
    Returns:
        str: wire string representation
    """
    assert_type(content_type, ContentType, "content_type")
    if content_type == ContentType.PROTOBUF:
        return 'application/x-protobuf'

    if content_type == ContentType.JSON:
        return 'application/json'

    raise NotImplementedError(
        "ContentType '{}' wire serialization not implemented".format(
            content_type.name))


def content_type_from_wire(string):
    """ Converts the ContentType wire string representation to the enum form.
    Args:
        string (str): wire string representation
    Returns:
        ContentType: enum value
    """
    assert_type(string, str, "string")
    if string == 'application/x-protobuf':
        return ContentType.PROTOBUF

    if string == 'application/json':
        return ContentType.JSON

    raise RuntimeError("Bad content_type {}".format(string))



class Context(object):
    def __init__(self, request, reply):
        self._request = request
        self._reply = reply
        self._addons = {}

    @property
    def request(self):
        return self._request

    @property
    def reply(self):
        return self._reply

    @property
    def addons(self):
        return self._addons



from is_wire.rpc.service_provider import ServiceProvider
from is_wire.rpc.interceptor import Interceptor
from is_wire.rpc.log_interceptor import LogInterceptor
from is_wire.rpc.tracing_interceptor import TracingInterceptor
from is_wire.rpc.metrics_interceptor import MetricsInterceptor

__all__ = [
    "ServiceProvider",
    "Interceptor",
    "LogInterceptor",
    "TracingInterceptor",
    "MetricsInterceptor",
]


from ..core import Channel, Subscription, Status, StatusCode, Logger
from ..core.utils import assert_type
from .context import Context
import traceback
import six
from google.protobuf.json_format import ParseError


class ServiceProvider(object):
    log = Logger("ServiceProvider", Logger.DEBUG)

    def __init__(self, channel):
        assert_type(channel, Channel, "channel")
        self._channel = channel
        self._services = {}
        self._interceptors_before = []
        self._interceptors_after = []
        self._subscriptions = []

    def delegate(self, topic, function, request_type, reply_type):
        """ Bind a function to a particular topic, so everytime a message is
            received in this topic the function will be called """
        assert_type(topic, six.string_types, "topic")
        if any(topic == s.name for s in self._subscriptions):
            raise RuntimeError(
                "Service on topic '{}' was already delegated".format(topic))

        self.log.debug("New service registered '{}'", topic)
        subscription = Subscription(self._channel, name=topic)
        self._subscriptions.append(subscription)
        wrapped = self.wrap(function, request_type, reply_type)
        self._services[subscription.id] = wrapped

    def add_interceptor(self, interceptor):
        """ Add an interceptor to the service provider. Interceptors provide
        a way to call functions before and after the actual service handler is
        called. For that the interceptor object passed must implement the
        Interceptor concept, that is, to have a before_call and after_call
        methods.
        """
        itype = type(interceptor)
        if not hasattr(itype, "before_call") and \
           not hasattr(itype, "after_call"):
            raise TypeError("Interceptors must implement the Interceptor"
                            "concept or derive from the Interceptor class")
        self._interceptors_before.append(interceptor.before_call)
        self._interceptors_after.append(interceptor.after_call)

    def should_serve(self, message):
        return message.subscription_id in self._services

    def serve(self, message):
        """ Attempts to serve the message. Raises runtime error if message
        cannot be served. Users can check if the message can be served by
        calling the should_serve method """
        try:
            service = self._services[message.subscription_id]
        except KeyError as error:
            why = "Cannot serve message with subscription_id='{}'".format(
                message.subscription_id)
            six.raise_from(RuntimeError(why), error)

        reply, timeouted = service(message)
        if reply.has_topic() and not timeouted:
            self._channel.publish(reply)

    def run(self):
        """ Blocks the current thread listening for requests """
        self.log.info("Listening for requests")
        while True:
            self.serve(self._channel.consume())

    def wrap(self, function, request_type, reply_type):
        def safe_call(*args):
            try:
                result = function(*args)
                assert_type(result, (Status, reply_type), "function result")
                return result
            except Exception:
                return Status(
                    code=StatusCode.INTERNAL_ERROR,
                    why="Service throwed exception:\n{}".format(
                        traceback.format_exc()),
                )

        def run_interceptors(interceptors, *args):
            for interceptor in interceptors:
                try:
                    interceptor(*args)
                except Exception:
                    trace = traceback.format_exc()
                    self.log.error("Interceptor throwed exception:\n{}", trace)

        def wrapper(request):
            reply = request.create_reply()
            context = Context(request, reply)

            run_interceptors(self._interceptors_before, context)

            if not request.deadline_exceeded():
                try:
                    arg = request.unpack(request_type)
                    result = safe_call(arg, context)
                    if isinstance(result, Status):
                        reply.status = result
                    else:
                        reply.pack(result)
                        reply.status = Status(code=StatusCode.OK)
                except ParseError:
                    why = "Expected request type '{}' but received something else"\
                        .format(request_type.DESCRIPTOR.full_name)
                    reply.status = Status(StatusCode.FAILED_PRECONDITION, why)
                except Exception:
                    trace = traceback.format_exc()
                    self.log.error("Unexpected error\n{}", trace)
                    reply.status = Status(StatusCode.INTERNAL_ERROR, trace)

            timeouted = request.deadline_exceeded()
            if timeouted:
                reply.status = Status(StatusCode.DEADLINE_EXCEEDED)

            run_interceptors(self._interceptors_after, context)

            return reply, timeouted

        return wrapper



from opencensus.trace import tracer
from ..utils import new_uuid
from opencensus.trace.span_context import SpanContext

# Patch opencensus to allow 64 bit trace ids
import opencensus.trace.span_context
import re
opencensus.trace.span_context.TRACE_ID_PATTERN = re.compile('[0-9a-f]{16}?')
opencensus.trace.span_context._INVALID_TRACE_ID = '0' * 16


class Tracer(object):
    def __init__(self, exporter=None, span_context=None):
        if span_context is None:
            span_context = SpanContext(trace_id=format(new_uuid(), '016x'))

        self.tracer = tracer.Tracer(
            exporter=exporter,
            span_context=span_context,
        )

    def span(self, name='span'):
        return self.tracer.span(name)

    def start_span(self, name='span'):
        return self.tracer.start_span(name)

    def end_span(self):
        return self.tracer.end_span()



from ..core import Logger, Tracer
from ..rpc import Interceptor


def service_name(context):
    return context.request.topic.split(".")[-1]


class TracingInterceptor(Interceptor):
    def __init__(self, exporter, span_namer=service_name):
        """ Construct a TracingInterceptor to automatically trace requests
        of a ServiceProvider.
        Args:
            exporter (opencensus.Exporter): object responsible for exporting
            spans to a tracer.
            span_namer (function: Context -> str): function responsible
            for generating the name of the request span.
        """
        self.log = Logger(name='TracingInterceptor')
        self.exporter = exporter
        self.namer = span_namer

    def before_call(self, context):
        self.tracer = Tracer(
            self.exporter, span_context=context.request.extract_tracing())
        context.addons["tracer"] = self.tracer
        self.span = self.tracer.start_span(name=self.namer(context))

    def after_call(self, context):
        status = context.reply.status
        if not status.ok():
            self.span.add_attribute("reply_to", context.request.reply_to)
            self.span.add_attribute("status_code", status.code.name)
            self.span.add_attribute("status_why", status.why)

        context.reply.inject_tracing(self.span)
        self.tracer.end_span()



from setuptools import setup, find_packages

setup(
    name='is_wire',
    version='1.1.4',
    description='',
    url='http://github.com/labviros/is-wire-py',
    author='labviros',
    license='MIT',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    zip_safe=False,
    install_requires=[
        'colorlog==3.1.4',
        'amqp==2.3.2',
        'enum34==1.1.6',
        'protobuf==3.6.0',
        'is-opencensus>=0.1.5.3',
        'prometheus-client==0.3.1',
    ]
    # change 'is-opencensus' to 'opencensus' when pull request accepted
)



from ..message import Message
from .content_type import content_type_from_wire, content_type_to_wire
from .status import Status, StatusCode
from . import wire_pb2
from google.protobuf import json_format
from six import binary_type


class WireV1(object):
    @staticmethod
    def from_amqp_message(amqp_message):
        message = Message()

        if not isinstance(amqp_message.body, binary_type):
            message.body = amqp_message.body.encode('latin')
        else:
            message.body = amqp_message.body

        delivery_info = amqp_message.delivery_info
        message.topic = delivery_info["routing_key"]
        message.subscription_id = delivery_info["consumer_tag"]

        properties = amqp_message.properties
        if "content_type" in properties:
            message.content_type = content_type_from_wire(
                properties["content_type"])

        if "correlation_id" in properties:
            message.correlation_id = int(properties["correlation_id"], 16)

        if "reply_to" in properties:
            message.reply_to = properties["reply_to"]

        if "expiration" in properties:
            message.timeout = int(properties["expiration"]) / 1000.0

        if "timestamp" in properties:
            message.created_at = properties["timestamp"] / 1000.0

        if "application_headers" in properties:
            if "rpc-status" in properties["application_headers"]:
                status = json_format.Parse(
                    properties["application_headers"]["rpc-status"],
                    wire_pb2.Status())
                message.status = Status(
                    code=StatusCode(status.code),
                    why=status.why,
                )
                del properties["application_headers"]["rpc-status"]

            message.metadata = properties["application_headers"]

        return message

    @staticmethod
    def to_amqp_properties(message):
        properties = {}
        properties["timestamp"] = int(message.created_at * 1000)

        if message.has_content_type():
            properties["content_type"] = content_type_to_wire(
                message.content_type)

        if message.has_correlation_id():
            properties["correlation_id"] = "{:X}".format(
                message.correlation_id)

        if message.has_reply_to():
            properties["reply_to"] = message.reply_to

        if message.has_timeout():
            properties["expiration"] = str(int(message.timeout * 1000))

        if len(message.metadata) != 0:
            properties["application_headers"] = message.metadata
        else:
            properties["application_headers"] = {}

        if message.has_status():
            status = wire_pb2.Status(
                code=message.status.code.value,
                why=message.status.why,
            )
            properties["application_headers"][
                "rpc-status"] = json_format.MessageToJson(
                    status, indent=0, including_default_value_fields=True)

        return properties



from is_wire.core.channel import Channel
from is_wire.core.subscription import Subscription
from is_wire.core.message import Message
from is_wire.core.logger import Logger
from is_wire.core.utils import now, new_uuid
from is_wire.core.wire.status import Status, StatusCode
from is_wire.core.wire.content_type import ContentType
from is_wire.core.tracing.tracer import Tracer
from opencensus.trace.exporters.zipkin_exporter import ZipkinExporter
from opencensus.trace.exporters.transports.background_thread import \
                                                     BackgroundThreadTransport

__all__ = [
    "Channel",
    "Subscription",
    "Message",
    "Logger",
    "now",
    "new_uuid",
    "Status",
    "StatusCode",
    "ContentType",
    "Tracer",
    "ZipkinExporter",
    "BackgroundThreadTransport",
]
